from django.db import models
from django.db.models.signals import post_save, pre_save
from django.contrib.auth.models import User
from exceptions import *
import tagging
from tagging.fields import TagField

import datetime
import difflib

class Page(models.Model):
    """
    Wiki pages are unique by title and slug.
    
    Pages are placeholder's for Revisions.  Each page can have many revisions.
    Upon creation a default hello world page is created and published.
    
    A Page can only have one unpublished Reversion.  A blank unpublished
    Revision is created when a Page is checked out.
    
    A Page cannot be checked in while an unblished Revision for it exists.
    """
    title = models.CharField(max_length=255, unique=True)
    slug = models.SlugField(unique=True)
    tags = TagField()
    creator = models.ForeignKey(User)
    is_editable = models.BooleanField(default=True)
    is_checked_out = models.BooleanField(default=False)
    created_on = models.DateTimeField(auto_now_add=True, editable=False)
    
    def check_out(self, user):
        """
        Before checking a page out, check if page is already checked out.  If
        the page is already checked out, raise PageAlreadyCheckedOut exception.
        
        If the page isn't checked out, create a skeleton revision, check out page.
        """
        if self.is_checked_out:
            raise PageAlreadyCheckedOut
        revision = Revision.objects.create(
            page=self,
            author=user,
            body=self.latest_revision().body
        )
        self.is_checked_out = True
        self.save()
        return revision

    def check_in(self):
        """
        When model is saved a pre_save signal will throw a UnpublishedRevisionExists
        exception if the unpublished revision isn't published or abandoned first.
        """
        self.is_checked_out = False
        self.save()
    
    def latest_revision(self):
        try:
            return Revision.objects.filter(page=self,is_published=True).latest('published_on')
        except Revision.DoesNotExist:
            return None

    def history(self):
        return Revision.objects.filter(page=self, is_published=True).order_by('-published_on')

    def unpublished_revision(self):
        try:
            return Revision.objects.filter(page=self, is_published=False).latest('created_on')
        except Revision.DoesNotExist:
            return None
    
    def count_revisions(self):
        return Revision.objects.filter(page=self, is_published=True).count()

    def who_checked_out(self):
        try:
            return Revision.objects.filter(page=self, is_published=False).latest('created_on').author
        except Revision.DoesNotExist:
            return None

    def compare(self, rev1, rev2):
        if rev1 == rev2:
            raise ComparingSameRevision
        revisions = Revision.objects.filter(page=self, pk__in=sorted([rev1, rev2]))
        if revisions.count() != 2:
            raise RevisionDoesNotExist

        comparison, created = Comparison.objects.get_or_create(
            page=self,
            rev1=revisions[0],
            rev2=revisions[1],
        )
        return comparison
        
    @models.permalink
    def get_absolute_url(self):
        return ('page_detail', (), { 'slug': self.slug })

    def __unicode__(self):
        return self.title   
    
    class Meta:
        ordering = ['title']

class Revision(models.Model):
    """
    All Revisions belong to a Page.
    
    A Revision is created only when a Page is checked out.
    
    A Revision can only be deleted before it is published, once published it
    cannot be changed.
    """
    page = models.ForeignKey(Page)
    author = models.ForeignKey(User)
    number = models.IntegerField(default=0, editable=False)
    body = models.TextField()
    is_published = models.BooleanField(default=False)
    published_on = models.DateTimeField(blank=True, null=True, editable=False)
    created_on = models.DateTimeField(auto_now_add=True, editable=False)
    edited_on = models.DateTimeField(auto_now=True, editable=False)

    def publish(self, check_in_page=True):
        self.is_published = True
        self.save()
        
        if check_in_page:
            self.page.check_in()

    def __unicode__(self):
        return '%s #%s' % (self.page.title, self.number)

    class Meta:
        ordering = ['number']
        unique_together = ("page", "number")

def create_first_revision(sender, instance, created, **kwargs):
    """Creates a "blank" revision for a new page."""
    if created:
        Revision.objects.create(
            page=instance,
            author=instance.creator,
            body="hello world!",
            is_published=True,
        )

def check_page_already_checked_out(page_new, page_old):
    """Checks to see if page is checked out, can't be checked out again."""
    if page_old is not None and page_new.is_checked_out and page_old.is_checked_out is True:
        raise PageAlreadyCheckedOut
   
def check_page_unpublished_revisions(page_new):
    """Before a page is checked in there can be no unpublished revisions"""
    if not page_new.is_checked_out and Revision.objects.filter(page=page_new, is_published=False).count() > 0:
        raise UnpublishedRevisionExists

def check_revision_already_published(sender, instance, **kwargs):
    """Published revisions can not be unpublished"""
    if instance.pk != None:
        if instance.is_published:
            try:
                Revision.objects.get(pk=instance.pk, is_published=True)
            except Revision.DoesNotExist:
                return
            raise AlreadyPublishedRevision
                
def update_published_on(sender, instance, **kwargs):
    """If page goes from not published to published, update published_on date."""
    published_on = datetime.datetime.now()
    latest = instance.page.latest_revision()

    if latest is None:
        number = 1
    else:
        number = latest.number + 1
        
    try:
        original = Revision.objects.get(pk=instance.pk)
        if not original.is_published and instance.is_published:
            instance.published_on = published_on
            instance.number = number
    except Revision.DoesNotExist:
        if instance.is_published:
            instance.published_on = published_on
            instance.number = number

def page_pre_save_maintenance(sender, instance, **kwargs):
    if instance.pk:
        check_page_already_checked_out(instance, Page.objects.get(pk=instance.pk))
        check_page_unpublished_revisions(instance)

post_save.connect(create_first_revision, sender=Page)
pre_save.connect(page_pre_save_maintenance, sender=Page)
pre_save.connect(check_revision_already_published, sender=Revision)
pre_save.connect(update_published_on, sender=Revision)

class Comparison(models.Model):
    page = models.ForeignKey(Page)
    rev1 = models.ForeignKey(Revision, related_name='rev1')
    rev2 = models.ForeignKey(Revision, related_name='rev2')
    diff_text = models.TextField(default='blank')
    
    def __unicode__(self):
        return u'%s: %s vs %s' % (self.page.title, self.rev1.pk, self.rev2.pk)

def comparison_calculate_diff(sender, instance, created, **kwargs):
    if created:
        from utils.diff_match_patch import diff_match_patch
        diff = diff_match_patch()
        diff_array = diff.diff_main(instance.rev1.body, instance.rev2.body)
        diff.diff_cleanupSemantic(diff_array)
        instance.diff_text = diff.diff_prettyHtml(diff_array)
        instance.save()
        
post_save.connect(comparison_calculate_diff, sender=Comparison)



import unittest
import random
from django.contrib.auth.models import User
from vz_wiki.models import Page, Revision
from vz_wiki.exceptions import *

TEST_SIZE = 50
random.seed()

class WikiTestCase(unittest.TestCase):
    def setUp(self):
        self.user = User.objects.create(
            username='wiki_admin',
            email='wiki_admin@localhost',
        )
        self.user2 = User.objects.create(
            username='wiki_user',
            email='wiki_user@localhost',
        )
        self.pages = []
        for x in range(TEST_SIZE):
            self.pages.append(Page.objects.create(
                title=u'test page number %s'%x,
                slug=u'test-page-number-%s'%x,
                tags='test',
                creator=self.user
            ))

    def testWiki(self):
        for y in range(TEST_SIZE/2):
            page1 = random.choice(self.pages)
            #self.assertEqual(page1.count_revisions(), 1)
            if page1.is_checked_out:
                self.failUnlessRaises(UnpublishedRevisionExists, page1.check_in)
                revivsion = page1.unpublished_revision()
                revivsion.publish()
                self.failUnlessRaises(AlreadyPublishedRevision, revivsion.publish)
                page1.check_in()
            page1.check_out(user=self.user2)
            self.failUnlessRaises(PageAlreadyCheckedOut, page1.check_out, self.user)
            page2 = random.choice(self.pages)



from django.contrib.auth.decorators import permission_required
from django.http import HttpResponseForbidden, HttpResponseBadRequest, Http404
from django.shortcuts import render_to_response, redirect, get_object_or_404
from django.template import RequestContext
from tagging.models import Tag, TaggedItem
from tagging.utils import parse_tag_input
from models import Page, Revision
from forms import PageForm, RevisionForm
from exceptions import RevisionDoesNotExist

def page_tags(request):
    tags_string = request.GET.get('tags', None)
    if tags_string is not None:
        tag_list = parse_tag_input(tags_string)
        page_list = TaggedItem.objects.get_union_by_model(Page, tags_string)
    else:
        page_list = None
        tag_list = None
    return render_to_response(
        'vz_wiki/page_tags.html',
        {
            'page_list': page_list,
            'tag_list': tag_list,
        },
        context_instance=RequestContext(request)
    )      

def compare_revisions(request, page_id):
    """
    Compare revisions from page using Python's Difflib.
    
    Page is page_id, revision number to compare are from GET.  Revision numbers
    are rev1 and rev2.
    
    Templates: ``compare_revisions.html``
    Context:
        page
            Page object
        comparision
            Comparision object from page, rev1 and rev2
    """
    page = get_object_or_404(Page, pk=page_id)
    rev1 = request.GET.get('rev1', None)
    rev2 = request.GET.get('rev2', None)
    if rev1 is None or rev2 is None:
        return HttpResponseBadRequest('Missing rev1 and rev2.')
    try:
        comparison = page.compare(rev1, rev2)
    except RevisionDoesNotExist:
        raise Http404
    return render_to_response(
        'vz_wiki/compare_revisions.html',
        {
            'page': page,
            'comparison': comparison,
        },
        context_instance=RequestContext(request)
    )     

def page_history(request, page_id):
    """
    List of page history.
    
    Templates: ``page_history.html``
    Context:
        page
            Page object
    """
    page = get_object_or_404(Page, pk=page_id)
    return render_to_response(
        'vz_wiki/page_history.html',
        { 'page': page, },
        context_instance=RequestContext(request)
    ) 
    

def abandon_revision(request, revision_id):
    """
    Abandons revision, deletes it without publishing it.
    
    Redirects to revision's parent page.
    
    Templates: none
    Context:
        none
    """
    revision = get_object_or_404(Revision, pk=revision_id, is_published=False)
    if revision.author != request.user:
        return HttpResponseForbidden('You cannot abandon this revision, it does not belong to you.')
    page = revision.page
    revision.delete()
    page.check_in()
    request.user.message_set.create(message='Revision abandoned.')
    return redirect(page)

abandon_revision = permission_required('vz_wiki.page.can_change')(abandon_revision)

def edit_page(request, page_id):
    """
    Creates/saves/publishes a page's revision.
    
    If page is checked out and current user isn't the "checker outer", don't allow
    edit.  Otherwise open the Page.unpublished_revision().
    
    Templates: ``edit_page.html``
    Context:
        page
            Page object
        form
            RevisionForm ojbect
    """
    page = get_object_or_404(Page, pk=page_id)
    if page.is_checked_out:
        if page.who_checked_out() != request.user:
            request.user.message_set.create(message='This page is already checked out.')
            return redirect(page)
        unpublished_revision = page.unpublished_revision()
    else:
        unpublished_revision = page.check_out(user=request.user)

    if request.method == 'POST':
        form = RevisionForm(request.POST, instance=unpublished_revision)
        if form.is_valid():
            Tag.objects.update_tags(page, form.cleaned_data['tags'])
            unpublished_revision = form.save(commit=False)
            if unpublished_revision.is_published:
                unpublished_revision.publish()
                request.user.message_set.create(message="Revision published.")
                return redirect(page)
            else:
                unpublished_revision.save()
    else:
        form = RevisionForm(
            initial={ 'tags': page.tags},
            instance=unpublished_revision,
        )

    return render_to_response(
        'vz_wiki/edit_page.html',
        {
            'form': form,
            'page': page,
            'unpublished_revision': unpublished_revision,
        },
        context_instance=RequestContext(request)
    )       
            
edit_page = permission_required('vz_wiki.page.can_change')(edit_page)    

def create_page(request):
    """
    Creates a new page.
    
    Templates: ``create_page.html``
    Context:
        form
            PageForm object
    """
    if request.method == 'POST':
        form = PageForm(request.POST)
        if form.is_valid():
            page = form.save(commit=False)
            page.creator = request.user
            page.save()
            return redirect(page)
    else:
        form = PageForm()

    return render_to_response(
        'vz_wiki/create_page.html',
        { 'form': form },
        context_instance=RequestContext(request)
    )
create_page = permission_required('vz_wiki.page.can_add')(create_page)



from django.forms import ModelForm
from tagging.forms import TagField
from models import Page, Revision

class PageForm(ModelForm):
    class Meta:
        model = Page
        exclude = ['creator', 'is_checked_out']

class RevisionForm(ModelForm):
    tags = TagField(label='Tags for Page')
    class Meta:
        model = Revision
        exclude = ['page', 'author']



from django.contrib import admin
from models import Page, Revision, Comparison

def latest_revision_display(obj):
    latest_revision = obj.latest_revision()
    if latest_revision is None:
        return '0'
    return u'%s' % latest_revision.number
latest_revision_display.short_description = 'Version'

def who_checked_out(obj):
    return obj.who_checked_out()
who_checked_out.short_description = 'Checked out by'

def make_editable(modelAdmin, request, queryset):
    queryset.update(is_editable=True)
make_editable.short_description = 'Make the selected pages editable'

def make_not_editable(modelAdmin, request, queryset):
    queryset.update(is_editable=False)
make_not_editable.short_description = 'Make the selected pages not editable'

def make_checked_in(modelAdmin, request, queryset):
    queryset.update(is_checked_out=False)
make_checked_in.short_description = 'Check in the selected pages'

def make_checked_out(modelAdmin, request, queryset):
    queryset.update(is_checked_out=True)
make_checked_out.short_description = 'Check out the selected pages'

class RevisionInline(admin.StackedInline):
    model = Revision
    extra = 1

class PageAdmin(admin.ModelAdmin):
    actions = [make_editable, make_not_editable, make_checked_in, make_checked_out]
    inlines = [RevisionInline]
    prepopulated_fields = {'slug': ('title',)}
    list_display = (
        'title', 
        latest_revision_display,
        'is_editable',
        'is_checked_out',
        who_checked_out,
        'created_on'
    )
    list_filter = ('is_editable', 'is_checked_out') 

admin.site.register(Page, PageAdmin)
admin.site.register(Revision)
admin.site.register(Comparison)



from django.conf.urls.defaults import *
from models import Page

urlpatterns = patterns('vz_wiki.views',
    url(r'^pages:create/$', 'create_page', name='create_page'),
    url(r'^pages:edit/(?P<page_id>\d+)/$', 'edit_page', name='edit_page'),
    url(r'^pages:abandon/(?P<revision_id>\d+)/$', 'abandon_revision', name='abandon_revision'),
    url(r'^pages:tags', 'page_tags', name='page_tags'),
    url(r'^pages:history/(?P<page_id>\d+)/$', 'page_history', name='page_history'),
    url(
        r'^pages:compare\-revisions/(?P<page_id>\d+)/$',
        'compare_revisions',
        name='compare_revisions'
    ),
)

urlpatterns += patterns('django.views.generic',
    url(
        r'^(?P<slug>[\w\-]+)/$',
        'list_detail.object_detail',
        {
            'queryset': Page.objects.select_related().all(),
            'template_name': 'vz_wiki/page_detail.html',
            'template_object_name': 'page',
        },
        name='page_detail',
    ),
    url(
        r'^pages:index/$',
        'list_detail.object_list',
        {
            'queryset': Page.objects.all(),
            'template_name': 'vz_wiki/page_list.html',
            'template_object_name': 'page',
        },
        name='page_list',
    ),
    url(
        r'^$',
        'simple.redirect_to',
        { 'url': '/wiki/index/' },
        name='index',
    ),
)



from django.conf import settings
from django import template
from django.template.defaultfilters import stringfilter

from BeautifulSoup import BeautifulSoup, Comment

import re

wiki_link_pattern = re.compile('\[\[([^\]]+)\]\]')
white_space_pattern = re.compile('[^\w]+')

register = template.Library()

@register.filter
@stringfilter
def wiki_link(text):
    """
    Search text for [[link_me]], replace with <a href="WIKI_BASE/link_me">link_me</a>
    """
    return wiki_link_pattern.sub(get_link, text)

wiki_link.is_safe = True

def get_link(match_obj):
    """
    Creates link text
    """
    text = match_obj.group(0)[2:-2]
    slug = white_space_pattern.sub('-', text.lower())
    return u'<a href="%s/%s" title="%s">%s</a>' % (settings.WIKI_BASE, slug, text, text)

@register.filter
def sanitize(value, allowed_tags=None):
    """
    Jacked from: http://www.djangosnippets.org/snippets/1655/
    
    Argument should be in form 'tag2:attr1:attr2 tag2:attr1 tag3', where tags
    are allowed HTML tags, and attrs are the allowed attributes for that tag.
    """
    js_regex = re.compile(r'[\s]*(&#x.{1,7})?'.join(list('javascript')))

    if allowed_tags is None:
        allowed_tags = settings.WIKI_ALLOWED_TAGS
    else:
        allowed_tags = '%s %s'%(allowed_tags, settings.WIKI_ALLOWED_TAGS)

    allowed_tags = [tag.split(':') for tag in allowed_tags.split()]
    allowed_tags = dict((tag[0], tag[1:]) for tag in allowed_tags)

    soup = BeautifulSoup(value)
    for comment in soup.findAll(text=lambda text: isinstance(text, Comment)):
        comment.extract()

    for tag in soup.findAll(True):
        if tag.name not in allowed_tags:
            tag.hidden = True
        else:
            tag.attrs = [(attr, js_regex.sub('', val)) for attr, val in tag.attrs
                         if attr in allowed_tags[tag.name]]

    return soup.renderContents().decode('utf8')



class PageAlreadyCheckedOut(Exception):
    "Paged check out, please check it in"
    pass

class UnpublishedRevisionExists(Exception):
    "This page has an unpublished revision, please publish or discard any drafts"
    pass

class AlreadyPublishedRevision(Exception):
    "This revision has already been published, this cannot be undone."
    pass

class ComparingSameRevision(Exception):
    "Revision numbers being compared are equal and cannot be compared."
    pass

class RevisionDoesNotExist(Exception):
    "Cannot compare revisions because one or more revisions do(es) not exist."
    pass



from setuptools import setup, find_packages
setup(
    name = "django-vz-wiki",
    version=__import__('vz_wiki').__version__,
    packages=find_packages(exclude=['ez_setup']),
    author = "Joe Vasquez",
    author_email = "joe.vasquez@gmail.com",
    license='BSD',
    description = "A Django powered wiki app",
    long_description=open('README.markdown').read(),
    url = "http://jobscry.net/projects/#vz_wiki",
    download_url = "http://github.com/jobscry/vz-wiki",
    include_package_data = True
)
